/**
 * Open Command Prompt Shell Extension
 * Copyright (C) Kai Liu.  All rights reserved.
 *
 * Please refer to readme.txt for information about this source code.
 * Please refer to license.txt for details about distribution and modification.
 **/

#include "CCmdOpen.hpp"
#include "RegHelpers.h"

STDMETHODIMP CCmdOpen::QueryInterface( REFIID riid, LPVOID *ppv )
{
	if (IsEqualIID(riid, IID_IUnknown))
	{
		*ppv = this;
	}
	else if (IsEqualIID(riid, IID_IShellExtInit))
	{
		*ppv = (LPSHELLEXTINIT)this;
	}
	else if (IsEqualIID(riid, IID_IContextMenu))
	{
		*ppv = (LPCONTEXTMENU)this;
	}
	else
	{
		*ppv = NULL;
		return(E_NOINTERFACE);
	}

	AddRef();
	return(S_OK);
}

STDMETHODIMP CCmdOpen::Initialize( LPCITEMIDLIST pidlFolder, LPDATAOBJECT pdtobj, HKEY hkeyProgID )
{
	// We need to check the TypeID values to determine which handler we are
	// initializing for...

	DWORD dwClassTypeID;

	if (!hkeyProgID) return(E_INVALIDARG);

	HKEY hkeyHandler;
	TCHAR szSubKey[64];
	wnsprintf(szSubKey, countof(szSubKey), TEXT("ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen);

	if (RegOpenKeyEx(hkeyProgID, szSubKey, 0, KEY_QUERY_VALUE, &hkeyHandler) == ERROR_SUCCESS)
	{
		RegGetDW(hkeyHandler, TEXT("TypeID"), &dwClassTypeID);
		RegCloseKey(hkeyHandler);
	}
	else return(E_INVALIDARG);


	// We'll be needing a buffer, and let's double it just to be safe
	TCHAR szPath[MAX_PATH << 1];

	// Make sure that we are working with a fresh list
	SLDestroy(m_hList);
	m_hList = SLCreate();

	// Background: just translate the folder ID
	if (dwClassTypeID == CLSTYPE_ID_CmdOpenBG)
	{
		if (pidlFolder && SHGetPathFromIDList(pidlFolder, szPath))
			SLAddString(m_hList, szPath);
	}

	// My Documents: just look up the folder path
	else if (dwClassTypeID == CLSTYPE_ID_CmdOpenMyDocs)
	{
		if (SHGetSpecialFolderPath(NULL, szPath, CSIDL_PERSONAL, FALSE))
			SLAddString(m_hList, szPath);
	}

	// Normal: convert pdtobj to hDrop and then enumerate
	else if (dwClassTypeID == CLSTYPE_ID_CmdOpenFolder)
	{
		FORMATETC format = { CF_HDROP, NULL, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
		STGMEDIUM medium;

		if (!pdtobj || pdtobj->GetData(&format, &medium) != S_OK)
			return(E_INVALIDARG);

		if (HDROP hDrop = (HDROP)GlobalLock(medium.hGlobal))
		{
			UINT uDrops = DragQueryFile(hDrop, -1, NULL, 0);

			for (UINT uDrop = 0; uDrop < uDrops; ++uDrop)
			{
				if ( DragQueryFile(hDrop, uDrop, szPath, countof(szPath)) &&
				     GetFileAttributes(szPath) & FILE_ATTRIBUTE_DIRECTORY )
				{
					SLAddString(m_hList, szPath);
				}
			}

			GlobalUnlock(medium.hGlobal);
		}

		ReleaseStgMedium(&medium);
	}


	// If there was any failure, the list would be empty...
	return((SLCheck(m_hList)) ? S_OK : E_INVALIDARG);
}

STDMETHODIMP CCmdOpen::QueryContextMenu( HMENU hmenu, UINT indexMenu, UINT idCmdFirst, UINT idCmdLast, UINT uFlags )
{
	if (uFlags & (CMF_DEFAULTONLY | CMF_NOVERBS))
		return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 0));

	// Load the localized menu text
	TCHAR szMenuText[MAX_MENUTEXT];
	LoadString(g_hModThisDll, IDS_MENUTEXT, szMenuText, countof(szMenuText));

	if (InsertMenu(hmenu, indexMenu, MF_STRING | MF_BYPOSITION, idCmdFirst, szMenuText))
		return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 1));

	return(MAKE_HRESULT(SEVERITY_SUCCESS, FACILITY_NULL, 0));
}

STDMETHODIMP CCmdOpen::InvokeCommand( LPCMINVOKECOMMANDINFO pici )
{
	// Ignore string verbs (high word must be zero)
	// The only valid command index is 0 (low word must be zero)
	if (pici->lpVerb)
		return(E_INVALIDARG);

	// Get the path to cmd.exe
	TCHAR szCommand[MAX_PATH + 0x10];
	LPTSTR lpszCommandAppend = szCommand;

	// To be safe, quote the path
	szCommand[0] = TEXT('"');
	lpszCommandAppend += 1;

	// Get the system path
	UINT cchSysDir = GetSystemDirectory(lpszCommandAppend, MAX_PATH);
	lpszCommandAppend += cchSysDir;

	// Append "\cmd.exe" and close the quotes; fall back to just "cmd.exe" if
	// for some strange reason, GetSystemDirectory failed
	if (cchSysDir && cchSysDir < MAX_PATH)
		SSStaticCpy(lpszCommandAppend, TEXT("\\cmd.exe\""));
	else
		SSStaticCpy(szCommand, TEXT("cmd.exe"));

	// Load the localized command prompt title
	TCHAR szCmdTitle[MAX_CMDTITLE];
	LoadString(g_hModThisDll, IDS_CMDTITLE, szCmdTitle, countof(szCmdTitle));

	// Initialize startup info
	PROCESS_INFORMATION pi;
	STARTUPINFO si;

	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	si.lpTitle = szCmdTitle;

	// Now for the trivial part...
	LPCTSTR lpszPath;

	while (lpszPath = (LPCTSTR)SLGetDataAndStep(m_hList))
	{
		BOOL bSuccess = CreateProcess(
			NULL,
			szCommand,
			NULL,
			NULL,
			FALSE,
			CREATE_DEFAULT_ERROR_MODE | CREATE_NEW_CONSOLE,
			NULL,
			lpszPath,
			&si,
			&pi
		);

		if (bSuccess)
		{
			CloseHandle(pi.hProcess);
			CloseHandle(pi.hThread);
		}
	}

	SLDestroy(m_hList);
	m_hList = NULL;

	return(S_OK);
}

STDMETHODIMP CCmdOpen::GetCommandString( UINT_PTR idCmd, UINT uFlags, UINT *pwReserved, LPSTR pszName, UINT cchMax )
{
	// In Vista, the build-in command has a verb of "cmd", so if we match
	// Vista's verb, then Vista's built-in facility will override ours when
	// it is active; this will eliminate duplicate menu entries.
	static const  CHAR szVerbA[] =  "cmd";
	static const WCHAR szVerbW[] = L"cmd";

	if (idCmd != 0 || cchMax < countof(szVerbW))
		return(E_INVALIDARG);

	switch (uFlags)
	{
		// The help text (status bar text) should not contain any of the
		// characters added for the menu access keys.

		case GCS_HELPTEXTA:
		{
			LoadStringA(g_hModThisDll, IDS_MENUTEXT, (LPSTR)pszName, cchMax);

			LPSTR lpszSrcA = (LPSTR)pszName;
			LPSTR lpszDestA = (LPSTR)pszName;

			while (*lpszSrcA && *lpszSrcA != '(')
			{
				if (*lpszSrcA != '&')
				{
					*lpszDestA = *lpszSrcA;
					++lpszDestA;
				}

				++lpszSrcA;
			}

			*lpszDestA = 0;
			return(S_OK);
		}

		case GCS_HELPTEXTW:
		{
			LoadStringW(g_hModThisDll, IDS_MENUTEXT, (LPWSTR)pszName, cchMax);

			LPWSTR lpszSrcW = (LPWSTR)pszName;
			LPWSTR lpszDestW = (LPWSTR)pszName;

			while (*lpszSrcW && *lpszSrcW != L'(')
			{
				if (*lpszSrcW != L'&')
				{
					*lpszDestW = *lpszSrcW;
					++lpszDestW;
				}

				++lpszSrcW;
			}

			*lpszDestW = 0;
			return(S_OK);
		}

		case GCS_VERBA:
		{
			SSStaticCpyA((LPSTR)pszName, szVerbA);
			return(S_OK);
		}

		case GCS_VERBW:
		{
			SSStaticCpyW((LPWSTR)pszName, szVerbW);
			return(S_OK);
		}
	}

	return(E_INVALIDARG);
}
