/**
 * Open Command Prompt Shell Extension
 * Copyright (C) Kai Liu.  All rights reserved.
 *
 * Please refer to readme.txt for information about this source code.
 * Please refer to license.txt for details about distribution and modification.
 **/

#include "globals.h"
#include "CCmdOpen.hpp"
#include "CCmdOpenClassFactory.hpp"
#include "RegHelpers.h"
#include "libs\Wow64.h"

// Bookkeeping globals (declared as extern in globals.h)
HMODULE g_hModThisDll;
CREF g_cRefThisDll;

// Prototypes for the self-registration/install/uninstall helper functions
STDAPI DllRegisterServerEx( LPCTSTR );
HRESULT Install( BOOL );
HRESULT Uninstall( );
BOOL WINAPI InstallFile( LPCTSTR, LPTSTR, LPTSTR );


#if defined(_USRDLL) && defined(_DLL)
#pragma comment(linker, "/entry:DllMain")
#endif
extern "C" BOOL WINAPI DllMain( HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved )
{
	switch (dwReason)
	{
		case DLL_PROCESS_ATTACH:
			g_hModThisDll = hInstance;
			g_cRefThisDll = 0;
			#ifndef _WIN64
			if (LOBYTE(LOWORD(GetVersion())) < 5) return(FALSE);
			#endif
			#ifdef _DLL
			DisableThreadLibraryCalls(hInstance);
			#endif
			break;

		case DLL_PROCESS_DETACH:
			break;

		case DLL_THREAD_ATTACH:
			break;

		case DLL_THREAD_DETACH:
			break;
	}

	return(TRUE);
}

STDAPI DllCanUnloadNow( )
{
	return((g_cRefThisDll == 0) ? S_OK : S_FALSE);
}

STDAPI DllGetClassObject( REFCLSID rclsid, REFIID riid, LPVOID *ppv )
{
	*ppv = NULL;

	if (IsEqualIID(rclsid, CLSID_CmdOpen))
	{
		LPCCMDOPENCLASSFACTORY lpCmdOpenClassFactory = new CCmdOpenClassFactory;
		if (lpCmdOpenClassFactory == NULL) return(E_OUTOFMEMORY);

		HRESULT hr = lpCmdOpenClassFactory->QueryInterface(riid, ppv);
		lpCmdOpenClassFactory->Release();
		return(hr);
	}

	return(CLASS_E_CLASSNOTAVAILABLE);
}

STDAPI DllRegisterServerEx( LPCTSTR lpszModuleName )
{
	HKEY hKey;

	// Register class
	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s"), CLSID_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSNAME_STR_CmdOpen);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s\\InprocServer32"), CLSID_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, lpszModuleName);
		RegSetSZ(hKey, TEXT("ThreadingModel"), TEXT("Apartment"));
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	// Register handlers
	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Directory\\Background\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetDW(hKey, TEXT("TypeID"), CLSTYPE_ID_CmdOpenBG);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Directory\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetDW(hKey, TEXT("TypeID"), CLSTYPE_ID_CmdOpenFolder);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("Drive\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
	{
		RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
		RegSetDW(hKey, TEXT("TypeID"), CLSTYPE_ID_CmdOpenFolder);
		RegCloseKey(hKey);
	} else return(SELFREG_E_CLASS);

	if (LOBYTE(LOWORD(GetVersion())) < 6)
	{
		// [Bug 18] Do not register the "My Documents" key for NT 6.x or newer
		if (hKey = RegOpen(HKEY_CLASSES_ROOT, TEXT("CLSID\\{450D8FBA-AD25-11D0-98A8-0800361B1103}\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
		{
			RegSetSZ(hKey, NULL, CLSID_STR_CmdOpen);
			RegSetDW(hKey, TEXT("TypeID"), CLSTYPE_ID_CmdOpenMyDocs);
			RegCloseKey(hKey);
		}
	}

	// Register approval
	if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Shell Extensions\\Approved"), NULL))
	{
		RegSetSZ(hKey, CLSID_STR_CmdOpen, CLSNAME_STR_CmdOpen);
		RegCloseKey(hKey);
	}

	return(S_OK);
}

STDAPI DllRegisterServer( )
{
	TCHAR szCurrentDllPath[MAX_PATH << 1];
	GetModuleFileName(g_hModThisDll, szCurrentDllPath, countof(szCurrentDllPath));
	return(DllRegisterServerEx(szCurrentDllPath));
}

STDAPI DllUnregisterServer( )
{
	HKEY hKey;
	BOOL bClassRemoved = TRUE;
	BOOL bApprovalRemoved = FALSE;

	// Unregister class
	if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("CLSID\\%s"), CLSID_STR_CmdOpen))
		bClassRemoved = FALSE;

	// Unregister handlers
	if (!Wow64CheckProcess())
	{
		/**
		 * Registry reflection sucks; it means that if we try to unregister the
		 * Wow64 extension, we'll also unregister the Win64 extension; the API
		 * to disable reflection seems to only affect changes in value, not key
		 * removals. :( This hack will disable the deletion of certain HKCR
		 * keys in the case of 32-on-64, and it should be pretty safe--unless
		 * the user had installed only the 32-bit extension without the 64-bit
		 * extension on Win64 (which should be a very rare scenario), there
		 * should be no undesirable effects to using this hack.
		 **/

		if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Directory\\Background\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
			bClassRemoved = FALSE;

		if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Directory\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
			bClassRemoved = FALSE;

		if (!RegDelete(HKEY_CLASSES_ROOT, TEXT("Drive\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen))
			bClassRemoved = FALSE;
	}

	// Any failures with the following key should be ignored
	RegDelete(HKEY_CLASSES_ROOT, TEXT("CLSID\\{450D8FBA-AD25-11D0-98A8-0800361B1103}\\ShellEx\\ContextMenuHandlers\\%s"), CLSNAME_STR_CmdOpen);

	// Unregister approval
	if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Shell Extensions\\Approved"), NULL))
	{
		LONG lResult = RegDeleteValue(hKey, CLSID_STR_CmdOpen);
		bApprovalRemoved = (lResult == ERROR_SUCCESS || lResult == ERROR_FILE_NOT_FOUND);
		RegCloseKey(hKey);
	}

	if (!bClassRemoved) return(SELFREG_E_CLASS);
	if (!bApprovalRemoved) return(S_FALSE);

	return(S_OK);
}

STDAPI DllInstall( BOOL bInstall, LPCWSTR pszCmdLine )
{
	// To install with bCopyFile=true
	// regsvr32.exe /i /n CmdOpen.dll
	//
	// To install with bCopyFile=false
	// regsvr32.exe /i:"NoCopy" /n CmdOpen.dll
	//
	// To uninstall
	// regsvr32.exe /u /i /n CmdOpen.dll
	//
	// DllInstall can also be invoked from a RegisterDlls INF section or from
	// a UnregisterDlls INF section, if the registration flags are set to 2.
	// Consult the documentation for RegisterDlls/UnregisterDlls for details.

	return( (bInstall) ?
		Install(pszCmdLine == NULL || StrCmpIW(pszCmdLine, L"NoCopy")) :
		Uninstall()
	);
}

HRESULT Install( BOOL bCopyFile )
{
	TCHAR szCurrentDllPath[MAX_PATH << 1];
	GetModuleFileName(g_hModThisDll, szCurrentDllPath, countof(szCurrentDllPath));

	TCHAR szSysDir[MAX_PATH + 0x20];
	UINT uSize = GetSystemDirectory(szSysDir, MAX_PATH);

	if (uSize && uSize < MAX_PATH)
	{
		LPTSTR lpszPath = szSysDir;
		LPTSTR lpszPathAppend = lpszPath + uSize;

		if (*(lpszPathAppend - 1) != TEXT('\\'))
			*lpszPathAppend++ = TEXT('\\');

		LPTSTR lpszTargetPath = (bCopyFile) ? lpszPath : szCurrentDllPath;

		if ( (!bCopyFile || InstallFile(szCurrentDllPath, lpszTargetPath, lpszPathAppend)) &&
		     DllRegisterServerEx(lpszTargetPath) == S_OK )
		{
			HKEY hKey;

			// DisableUNCCheck (KB156276)
			if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Command Processor"), NULL))
			{
				RegSetDW(hKey, TEXT("DisableUNCCheck"), 1);
				RegCloseKey(hKey);
			}

			// Uninstaller entries
			RegDelete(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen);

			if (hKey = RegOpen(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen))
			{
				TCHAR szUninstall[MAX_PATH << 1];
				wnsprintf(szUninstall, countof(szUninstall), TEXT("regsvr32.exe /u /i /n \"%s\""), lpszTargetPath);

				static const TCHAR szURLFull[] = TEXT("http://www.ktechcomputing.com/cmdopen/");
				TCHAR szURLBase[countof(szURLFull)];
				SSStaticCpy(szURLBase, szURLFull);
				szURLBase[30] = 0; // strlen("http://www.ktechcomputing.com/")

				SSStaticCpy(lpszPathAppend, TEXT("cmd.exe"));

				RegSetSZ(hKey, TEXT("DisplayIcon"), lpszPath);
				RegSetSZ(hKey, TEXT("DisplayName"), TEXT(CMDOPEN_NAME_STR));
				RegSetSZ(hKey, TEXT("DisplayVersion"), TEXT(CMDOPEN_VERSION_STR));
				RegSetSZ(hKey, TEXT("HelpLink"), szURLFull);
				RegSetDW(hKey, TEXT("NoModify"), 1);
				RegSetDW(hKey, TEXT("NoRepair"), 1);
				RegSetSZ(hKey, TEXT("Publisher"), TEXT("Kai Liu"));
				RegSetSZ(hKey, TEXT("UninstallString"), szUninstall);
				RegSetSZ(hKey, TEXT("URLInfoAbout"), szURLBase);
				RegCloseKey(hKey);
			}

			return(S_OK);

		} // if copied & registered

	} // if valid sysdir

	return(E_FAIL);
}

HRESULT Uninstall( )
{
	HRESULT hr = S_OK;

	// Schedule the DLL to be deleted at shutdown/reboot
	TCHAR szCurrentDllPath[MAX_PATH << 1];
	GetModuleFileName(g_hModThisDll, szCurrentDllPath, countof(szCurrentDllPath));
	if (!MoveFileEx(szCurrentDllPath, NULL, MOVEFILE_DELAY_UNTIL_REBOOT)) hr = E_FAIL;

	// Unregister
	if (DllUnregisterServer() != S_OK) hr = E_FAIL;

	// We don't need the uninstall strings any more...
	RegDelete(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s"), CLSNAME_STR_CmdOpen);

	return(hr);
}

BOOL WINAPI InstallFile( LPCTSTR lpszSource, LPTSTR lpszDest, LPTSTR lpszDestAppend )
{
	static const TCHAR szShellExt[] = TEXT("ShellExt");
	static const TCHAR szDestFile[] = TEXT("\\") TEXT(CMDOPEN_FILENAME_STR);

	SSStaticCpy(lpszDestAppend, szShellExt);
	lpszDestAppend += countof(szShellExt) - 1;

	// Create directory if necessary
	if (GetFileAttributes(lpszDest) == INVALID_FILE_ATTRIBUTES)
		CreateDirectory(lpszDest, NULL);

	SSStaticCpy(lpszDestAppend, szDestFile);
	lpszDestAppend += countof(szDestFile) - 1;

	// No need to copy if the source and destination are the same
	if (StrCmpI(lpszSource, lpszDest) == 0)
		return(TRUE);

	// If the destination file does not already exist, just copy
	if (GetFileAttributes(lpszDest) == INVALID_FILE_ATTRIBUTES)
		return(CopyFile(lpszSource, lpszDest, FALSE));

	// If destination file exists and cannot be overwritten
	TCHAR szTemp[MAX_PATH + 0x20];
	SIZE_T cbDest = BYTEDIFF(lpszDestAppend, lpszDest);
	LPTSTR lpszTempAppend = (LPTSTR)BYTEADD(szTemp, cbDest);

	memcpy(szTemp, lpszDest, cbDest);
	*lpszTempAppend++ = TEXT('.');
	*(lpszTempAppend + 1) = 0;

	for (TCHAR ch = TEXT('0'); ch <= TEXT('9'); ++ch)
	{
		if (CopyFile(lpszSource, lpszDest, FALSE))
			return(TRUE);

		*lpszTempAppend = ch;

		if (MoveFileEx(lpszDest, szTemp, MOVEFILE_REPLACE_EXISTING))
			MoveFileEx(szTemp, NULL, MOVEFILE_DELAY_UNTIL_REBOOT);
	}

	return(FALSE);
}
